#include "functions.h"
#include <fcntl.h>
#include <string.h>
#include <sys/dir.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdio.h>

void copyFile(char * filename, int sockfd, int mv) {
	int fd;
	char buf[FRAGSIZE];
	int n;
	char c = SPCCHR;
	
	strcpy(buf, filename);
	write(sockfd, buf, strlen(buf));
	read(sockfd, buf, 1);
	
	fd = open(filename, O_RDONLY, 0);
	do {
		n = read(fd, buf, FRAGSIZE);
		if (n != 0) write(sockfd, buf, n); else write(sockfd, &c, 1);
	} while (n == FRAGSIZE);
	close(fd);
	read(sockfd, &c, 1);
	if (mv) remove(filename);
}

int countFiles(char *dirname) {
	int i = 0;
	struct dirent *dp;
	DIR *dfd;
	
	if ((dfd = opendir(dirname)) != NULL) {
		while ((dp = readdir(dfd)) != NULL) {
			if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
			i++;
		}
		closedir(dfd);
	}
	
	return i;
}

void copyNodeDirectory(char * dirname, int sockfd, int mv) {
	char name[MAX_PATH];
	struct dirent *dp;
	DIR *dfd;
	int i;
	
	i = countFiles(dirname);
	write(sockfd, dirname, strlen(dirname));
	read(sockfd, name, 1);
	write(sockfd, &i, 1);
	
	if ((dfd = opendir(dirname)) != NULL) {
		while ((dp = readdir(dfd)) != NULL) {
			if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
			if (strlen(dirname)+strlen(dp->d_name)+2 <= sizeof(name)) {
				sprintf(name, "%s/%s", dirname, dp->d_name);
				copyFile(name, sockfd, mv);
			}
		}
		closedir(dfd);
	}
	if (mv) rmdir(dirname);
}

void copyDirectory(char * dirname, int sockfd, int mv) {
	char name[MAX_PATH];
	struct dirent *dp;
	DIR *dfd;
	int i;
	char tmpn = 'n', tmpf = 'f', tmpd = 'd';
	char * tmp[10];
	
	i = countFiles(dirname);
	write(sockfd, dirname, strlen(dirname));
	read(sockfd, name, 1);
	write(sockfd, &i, 1);
	
	if ((dfd = opendir(dirname)) != NULL) {
		while ((dp = readdir(dfd)) != NULL) {
			if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
			if (strlen(dirname)+strlen(dp->d_name)+2 <= sizeof(name)) {
				sprintf(name, "%s/%s", dirname, dp->d_name);
				if (!isDirectory(name)) {
					write(sockfd, &tmpf, 1);
					read(sockfd, &tmp, 1);
					copyFile(name, sockfd, mv);
				} else {
					if (isNodeDirectory(name)) {
						write(sockfd, &tmpn, 1);
						read(sockfd, &tmp, 1);
						copyNodeDirectory(name, sockfd, mv);
					} else {
						write(sockfd, &tmpd, 1);
						read(sockfd, &tmp, 1);
						copyDirectory(name, sockfd, mv);
					}
				}
			}
		}
		closedir(dfd);
	}
	if (mv) rmdir(dirname);
}

int isDirectory(char * name) {
	DIR *dfd;
	int result = 0;
	
	if ((dfd = opendir(name)) != NULL) {
		result = 1;
		closedir(dfd);
	}
	
	return result;
}

int isNodeDirectory(char * dirname) {
	char name[MAX_PATH];
	struct dirent *dp;
	DIR *dfd;
	int result = 1;
	
	if ((dfd = opendir(dirname)) != NULL) {
		while ((dp = readdir(dfd)) != NULL && result) {
			if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
			if (strlen(dirname)+strlen(dp->d_name)+2 <= sizeof(name)) {
				sprintf(name, "%s/%s", dirname, dp->d_name);
				if (isDirectory(name)) result = 0;
			}
		}
		closedir(dfd);
	}
	
	return result;
}

void receiveFile(int sockfd) {
	char buf[FRAGSIZE];
	char str[FRAGSIZE];
	int n;
	int fd;
	int rcvf = 1;
	
	n = read(sockfd, buf, FRAGSIZE);
	write(sockfd, buf, 1);
	strncpy(str, buf, n);
	str[n] = '\0';
	fd = creat(str, 0744);
	while (rcvf) {
		if ((n = read(sockfd, buf, FRAGSIZE)) == FRAGSIZE) {
			write(fd, buf, n);	
		} else {
			if (buf[0] != SPCCHR) write(fd, buf, n);
			rcvf = 0;
		}
	}
	close(fd);
	write(sockfd, buf, 1);
}

void receiveNodeDirectory(int sockfd) {
	char buf[FRAGSIZE];
	int i, n;
	
	n = read(sockfd, buf, FRAGSIZE);
	write(sockfd, buf, 1);
	buf[n] = '\0';
	mkdir(buf, 0755);
	read(sockfd, &n, 1);
	for (i=0; i<n; i++) {
		receiveFile(sockfd);
	}
}

void receiveDirectory(int sockfd) {
	char buf[FRAGSIZE];
	int i, n;
	char c;
	
	n = read(sockfd, buf, FRAGSIZE);
	write(sockfd, buf, 1);
	buf[n] = '\0';
	mkdir(buf, 0755);
	read(sockfd, &n, 1);
	for (i=0; i<n; i++) {
		read(sockfd, &c, 1);
		write(sockfd, &c, 1);
		if (c == 'f') receiveFile(sockfd);
		else if (c == 'n') receiveNodeDirectory(sockfd);
		else if (c == 'd') receiveDirectory(sockfd);
	}
}

